@@ -28,8 +28,8 @@
 # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-from __future__ import division, print_function, absolute_import
 import warnings
+import operator
 
 import numpy
 from . import _ni_support
@@ -132,7 +132,7 @@ def generate_binary_structure(rank, connectivity):
          will be applied, as returned by `np.ndim`.
     connectivity : int
          `connectivity` determines which elements of the output array belong
-         to the structure, i.e. are considered as neighbors of the central
+         to the structure, i.e., are considered as neighbors of the central
          element. Elements up to a squared distance of `connectivity` from
          the center are considered neighbors. `connectivity` may range from 1
          (no diagonal elements are neighbors) to `rank` (all elements are
@@ -151,9 +151,9 @@ def generate_binary_structure(rank, connectivity):
     Notes
     -----
     `generate_binary_structure` can only create structuring elements with
-    dimensions equal to 3, i.e. minimal dimensions. For larger structuring
-    elements, that are useful e.g. for eroding large objects, one may either
-    use   `iterate_structure`, or create directly custom arrays with
+    dimensions equal to 3, i.e., minimal dimensions. For larger structuring
+    elements, that are useful e.g., for eroding large objects, one may either
+    use `iterate_structure`, or create directly custom arrays with
     numpy functions such as `numpy.ones`.
 
     Examples
@@ -214,6 +214,11 @@ def generate_binary_structure(rank, connectivity):
 
 def _binary_erosion(input, structure, iterations, mask, output,
                     border_value, origin, invert, brute_force):
+    try:
+        iterations = operator.index(iterations)
+    except TypeError as e:
+        raise TypeError('iterations parameter should be an integer') from e
+
     input = numpy.asarray(input)
     if numpy.iscomplexobj(input):
         raise TypeError('Complex type not supported')
@@ -225,7 +230,7 @@ def _binary_erosion(input, structure, iterations, mask, output,
         raise RuntimeError('structure and input must have same dimensionality')
     if not structure.flags.contiguous:
         structure = structure.copy()
-    if numpy.product(structure.shape, axis=0) < 1:
+    if numpy.prod(structure.shape, axis=0) < 1:
         raise RuntimeError('structure must not be empty')
     if mask is not None:
         mask = numpy.asarray(mask)
@@ -239,7 +244,11 @@ def _binary_erosion(input, structure, iterations, mask, output,
     else:
         output = bool
     output = _ni_support._get_output(output, input)
-
+    temp_needed = numpy.may_share_memory(input, output)
+    if temp_needed:
+        # input and output arrays cannot share memory
+        temp = output
+        output = _ni_support._get_output(output.dtype, input)
     if iterations == 1:
         _nd_image.binary_erosion(input, structure, mask, output,
                                  border_value, origin, invert, cit, 0)
@@ -260,7 +269,6 @@ def _binary_erosion(input, structure, iterations, mask, output,
             structure = structure.copy()
         _nd_image.binary_erosion2(output, structure, mask, iterations - 1,
                                   origin, invert, coordinate_list)
-        return output
     else:
         tmp_in = numpy.empty_like(input, dtype=bool)
         tmp_out = output
@@ -276,13 +284,16 @@ def _binary_erosion(input, structure, iterations, mask, output,
                 tmp_in, structure, mask, tmp_out,
                 border_value, origin, invert, cit, 0)
             ii += 1
-        return output
+    if temp_needed:
+        temp[...] = output
+        output = temp
+    return output
 
 
 def binary_erosion(input, structure=None, iterations=1, mask=None, output=None,
                    border_value=0, origin=0, brute_force=False):
     """
-    Multi-dimensional binary erosion with a given structuring element.
+    Multidimensional binary erosion with a given structuring element.
 
     Binary erosion is a mathematical morphology operation used for image
     processing.
@@ -296,7 +307,7 @@ def binary_erosion(input, structure=None, iterations=1, mask=None, output=None,
         Structuring element used for the erosion. Non-zero elements are
         considered True. If no structuring element is provided, an element
         is generated with a square connectivity equal to one.
-    iterations : {int, float}, optional
+    iterations : int, optional
         The erosion is repeated `iterations` times (one, by default).
         If iterations is less than 1, the erosion is repeated until the
         result does not change anymore.
@@ -380,7 +391,7 @@ def binary_dilation(input, structure=None, iterations=1, mask=None,
                     output=None, border_value=0, origin=0,
                     brute_force=False):
     """
-    Multi-dimensional binary dilation with the given structuring element.
+    Multidimensional binary dilation with the given structuring element.
 
     Parameters
     ----------
@@ -391,10 +402,11 @@ def binary_dilation(input, structure=None, iterations=1, mask=None,
         Structuring element used for the dilation. Non-zero elements are
         considered True. If no structuring element is provided an element
         is generated with a square connectivity equal to one.
-    iterations : {int, float}, optional
+    iterations : int, optional
         The dilation is repeated `iterations` times (one, by default).
         If iterations is less than 1, the dilation is repeated until the
-        result does not change anymore.
+        result does not change anymore. Only an integer of iterations is
+        accepted.
     mask : array_like, optional
         If a mask is given, only those elements with a True value at
         the corresponding mask element are modified at each iteration.
@@ -510,7 +522,7 @@ def binary_dilation(input, structure=None, iterations=1, mask=None,
 def binary_opening(input, structure=None, iterations=1, output=None,
                    origin=0, mask=None, border_value=0, brute_force=False):
     """
-    Multi-dimensional binary opening with the given structuring element.
+    Multidimensional binary opening with the given structuring element.
 
     The *opening* of an input image by a structuring element is the
     *dilation* of the *erosion* of the image by the structuring element.
@@ -526,11 +538,11 @@ def binary_opening(input, structure=None, iterations=1, output=None,
         is generated with a square connectivity equal to one (i.e., only
         nearest neighbors are connected to the center, diagonally-connected
         elements are not considered neighbors).
-    iterations : {int, float}, optional
+    iterations : int, optional
         The erosion step of the opening, then the dilation step are each
         repeated `iterations` times (one, by default). If `iterations` is
         less than 1, each operation is repeated until the result does
-        not change anymore.
+        not change anymore. Only an integer of iterations is accepted.
     output : ndarray, optional
         Array of the same shape as input, into which the output is placed.
         By default, a new array is created.
@@ -568,7 +580,7 @@ def binary_opening(input, structure=None, iterations=1, output=None,
     -----
     *Opening* [1]_ is a mathematical morphology operation [2]_ that
     consists in the succession of an erosion and a dilation of the
-    input with the same structuring element. Opening therefore removes
+    input with the same structuring element. Opening, therefore, removes
     objects smaller than the structuring element.
 
     Together with *closing* (`binary_closing`), opening can be used for
@@ -633,7 +645,7 @@ def binary_opening(input, structure=None, iterations=1, output=None,
 def binary_closing(input, structure=None, iterations=1, output=None,
                    origin=0, mask=None, border_value=0, brute_force=False):
     """
-    Multi-dimensional binary closing with the given structuring element.
+    Multidimensional binary closing with the given structuring element.
 
     The *closing* of an input image by a structuring element is the
     *erosion* of the *dilation* of the image by the structuring element.
@@ -649,11 +661,11 @@ def binary_closing(input, structure=None, iterations=1, output=None,
         is generated with a square connectivity equal to one (i.e., only
         nearest neighbors are connected to the center, diagonally-connected
         elements are not considered neighbors).
-    iterations : {int, float}, optional
+    iterations : int, optional
         The dilation step of the closing, then the erosion step are each
         repeated `iterations` times (one, by default). If iterations is
         less than 1, each operations is repeated until the result does
-        not change anymore.
+        not change anymore. Only an integer of iterations is accepted.
     output : ndarray, optional
         Array of the same shape as input, into which the output is placed.
         By default, a new array is created.
@@ -779,7 +791,7 @@ def binary_closing(input, structure=None, iterations=1, output=None,
 def binary_hit_or_miss(input, structure1=None, structure2=None,
                        output=None, origin1=0, origin2=None):
     """
-    Multi-dimensional binary hit-or-miss transform.
+    Multidimensional binary hit-or-miss transform.
 
     The hit-or-miss transform finds the locations of a given pattern
     inside the input image.
@@ -815,7 +827,7 @@ def binary_hit_or_miss(input, structure1=None, structure2=None,
 
     See also
     --------
-    ndimage.morphology, binary_erosion
+    binary_erosion
 
     References
     ----------
@@ -888,7 +900,7 @@ def binary_hit_or_miss(input, structure1=None, structure2=None,
 def binary_propagation(input, structure=None, mask=None,
                        output=None, border_value=0, origin=0):
     """
-    Multi-dimensional binary propagation with the given structuring element.
+    Multidimensional binary propagation with the given structuring element.
 
     Parameters
     ----------
@@ -1024,7 +1036,7 @@ def binary_fill_holes(input, structure=None, output=None, origin=0):
     Parameters
     ----------
     input : array_like
-        n-dimensional binary array with holes to be filled
+        N-D binary array with holes to be filled
     structure : array_like, optional
         Structuring element used in the computation; large-size elements
         make computations faster but may miss holes separated from the
@@ -1145,7 +1157,7 @@ def grey_erosion(input, size=None, footprint=None, structure=None,
     See also
     --------
     binary_erosion, grey_dilation, grey_opening, grey_closing
-    generate_binary_structure, ndimage.minimum_filter
+    generate_binary_structure, minimum_filter
 
     Notes
     -----
@@ -1204,7 +1216,7 @@ def grey_erosion(input, size=None, footprint=None, structure=None,
 
     """
     if size is None and footprint is None and structure is None:
-        raise ValueError("size, footprint or structure must be specified")
+        raise ValueError("size, footprint, or structure must be specified")
 
     return filters._min_or_max_filter(input, size, footprint, structure,
                                       output, mode, cval, origin, 1)
@@ -1255,7 +1267,7 @@ def grey_dilation(input, size=None, footprint=None, structure=None,
     See also
     --------
     binary_dilation, grey_erosion, grey_closing, grey_opening
-    generate_binary_structure, ndimage.maximum_filter
+    generate_binary_structure, maximum_filter
 
     Notes
     -----
@@ -1329,7 +1341,7 @@ def grey_dilation(input, size=None, footprint=None, structure=None,
 
     """
     if size is None and footprint is None and structure is None:
-        raise ValueError("size, footprint or structure must be specified")
+        raise ValueError("size, footprint, or structure must be specified")
     if structure is not None:
         structure = numpy.asarray(structure)
         structure = structure[tuple([slice(None, None, -1)] *
@@ -1361,10 +1373,10 @@ def grey_dilation(input, size=None, footprint=None, structure=None,
 def grey_opening(input, size=None, footprint=None, structure=None,
                  output=None, mode="reflect", cval=0.0, origin=0):
     """
-    Multi-dimensional greyscale opening.
+    Multidimensional grayscale opening.
 
-    A greyscale opening consists in the succession of a greyscale erosion,
-    and a greyscale dilation.
+    A grayscale opening consists in the succession of a grayscale erosion,
+    and a grayscale dilation.
 
     Parameters
     ----------
@@ -1444,10 +1456,10 @@ def grey_opening(input, size=None, footprint=None, structure=None,
 def grey_closing(input, size=None, footprint=None, structure=None,
                  output=None, mode="reflect", cval=0.0, origin=0):
     """
-    Multi-dimensional greyscale closing.
+    Multidimensional grayscale closing.
 
-    A greyscale closing consists in the succession of a greyscale dilation,
-    and a greyscale erosion.
+    A grayscale closing consists in the succession of a grayscale dilation,
+    and a grayscale erosion.
 
     Parameters
     ----------
@@ -1527,7 +1539,7 @@ def grey_closing(input, size=None, footprint=None, structure=None,
 def morphological_gradient(input, size=None, footprint=None, structure=None,
                            output=None, mode="reflect", cval=0.0, origin=0):
     """
-    Multi-dimensional morphological gradient.
+    Multidimensional morphological gradient.
 
     The morphological gradient is calculated as the difference between a
     dilation and an erosion of the input with a given structuring element.
@@ -1568,7 +1580,7 @@ def morphological_gradient(input, size=None, footprint=None, structure=None,
 
     See also
     --------
-    grey_dilation, grey_erosion, ndimage.gaussian_gradient_magnitude
+    grey_dilation, grey_erosion, gaussian_gradient_magnitude
 
     Notes
     -----
@@ -1641,7 +1653,7 @@ def morphological_laplace(input, size=None, footprint=None,
                           structure=None, output=None,
                           mode="reflect", cval=0.0, origin=0):
     """
-    Multi-dimensional morphological laplace.
+    Multidimensional morphological laplace.
 
     Parameters
     ----------
@@ -1691,7 +1703,7 @@ def morphological_laplace(input, size=None, footprint=None,
 def white_tophat(input, size=None, footprint=None, structure=None,
                  output=None, mode="reflect", cval=0.0, origin=0):
     """
-    Multi-dimensional white tophat filter.
+    Multidimensional white tophat filter.
 
     Parameters
     ----------
@@ -1749,7 +1761,7 @@ def black_tophat(input, size=None, footprint=None,
                  structure=None, output=None, mode="reflect",
                  cval=0.0, origin=0):
     """
-    Multi-dimensional black tophat filter.
+    Multidimensional black tophat filter.
 
     Parameters
     ----------
@@ -1815,45 +1827,49 @@ def distance_transform_bf(input, metric="euclidean", sampling=None,
 
     In addition to the distance transform, the feature transform can
     be calculated. In this case the index of the closest background
-    element is returned along the first axis of the result.
+    element to each foreground element is returned in a separate array.
 
     Parameters
     ----------
     input : array_like
         Input
-    metric : str, optional
-        Three types of distance metric are supported: 'euclidean', 'taxicab'
-        and 'chessboard'.
-    sampling : {int, sequence of ints}, optional
-        This parameter is only used in the case of the euclidean `metric`
-        distance transform.
-
-        The sampling along each axis can be given by the `sampling` parameter
-        which should be a sequence of length equal to the input rank, or a
-        single number in which the `sampling` is assumed to be equal along all
-        axes.
+    metric : {'euclidean', 'taxicab', 'chessboard'}, optional
+        'cityblock' and 'manhattan' are also valid, and map to 'taxicab'.
+        The default is 'euclidean'.
+    sampling : float, or sequence of float, optional
+        This parameter is only used when `metric` is 'euclidean'.
+        Spacing of elements along each dimension. If a sequence, must be of
+        length equal to the input rank; if a single number, this is used for
+        all axes. If not specified, a grid spacing of unity is implied.
     return_distances : bool, optional
-        The `return_distances` flag can be used to indicate if the distance
-        transform is returned.
-
-        The default is True.
+        Whether to calculate the distance transform.
+        Default is True.
     return_indices : bool, optional
-        The `return_indices` flags can be used to indicate if the feature
-        transform is returned.
-
-        The default is False.
-    distances : float64 ndarray, optional
-        Optional output array to hold distances (if `return_distances` is
-        True).
-    indices : int64 ndarray, optional
-        Optional output array to hold indices (if `return_indices` is True).
+        Whether to calculate the feature transform.
+        Default is False.
+    distances : ndarray, optional
+        An output array to store the calculated distance transform, instead of 
+        returning it.
+        This parameter is only used when `return_distances` is True.
+        It must be the same shape as `input`, and of type float64 if `metric`
+        is 'euclidean', uint32 otherwise.
+    indices : int32 ndarray, optional
+        An output array to store the calculated feature transform, instead of 
+        returning it.
+        Its shape must be `(input.ndim,) + input.shape`.
 
     Returns
     -------
-    distances : ndarray
-        Distance array if `return_distances` is True.
-    indices : ndarray
-        Indices array if `return_indices` is True.
+    distances : ndarray, optional
+        The calculated distance transform. Returned only when
+        `return_distances` is True and `distances` is not supplied.
+        It will have the same shape as the input array.
+    indices : int32 ndarray, optional
+        The calculated feature transform. It has an input-shaped array for each 
+        dimension of the input. See distance_transform_edt documentation for an 
+        example.
+        Returned only when `return_indices` is True and `indices` is not 
+        supplied.
 
     Notes
     -----
@@ -1944,12 +1960,16 @@ def distance_transform_cdt(input, metric='chessboard', return_distances=True,
                            return_indices=False, distances=None, indices=None):
     """
     Distance transform for chamfer type of transforms.
+    
+    In addition to the distance transform, the feature transform can
+    be calculated. In this case the index of the closest background
+    element to each foreground element is returned in a separate array.
 
     Parameters
     ----------
     input : array_like
         Input
-    metric : {'chessboard', 'taxicab'}, optional
+    metric : {'chessboard', 'taxicab'} or array_like, optional
         The `metric` determines the type of chamfering that is done. If the
         `metric` is equal to 'taxicab' a structure is generated using
         generate_binary_structure with a squared distance equal to 1. If
@@ -1958,22 +1978,38 @@ def distance_transform_cdt(input, metric='chessboard', return_distances=True,
         the dimensionality of the array. These choices correspond to the
         common interpretations of the 'taxicab' and the 'chessboard'
         distance metrics in two dimensions.
+        A custom metric may be provided, in the form of a matrix where
+        each dimension has a length of three.
+        'cityblock' and 'manhattan' are also valid, and map to 'taxicab'.
+        The default is 'chessboard'.
+    return_distances : bool, optional
+        Whether to calculate the distance transform.
+        Default is True.
+    return_indices : bool, optional
+        Whether to calculate the feature transform.
+        Default is False.
+    distances : int32 ndarray, optional
+        An output array to store the calculated distance transform, instead of 
+        returning it.
+        This parameter is only used when `return_distances` is True.
+        It must be the same shape as `input`.
+    indices : int32 ndarray, optional
+        An output array to store the calculated feature transform, instead of 
+        returning it.
+        Its shape must be `(input.ndim,) + input.shape`.
 
-        The default for `metric` is 'chessboard'.
-    return_distances, return_indices : bool, optional
-        The `return_distances`, and `return_indices` flags can be used to
-        indicate if the distance transform, the feature transform, or both
-        must be returned.
-
-        If the feature transform is returned (``return_indices=True``),
-        the index of the closest background element is returned along
-        the first axis of the result.
-
-        The `return_distances` default is True, and the
-        `return_indices` default is False.
-    distances, indices : ndarrays of int32, optional
-        The `distances` and `indices` arguments can be used to give optional
-        output arrays that must be the same shape as `input`.
+    Returns
+    -------
+    distances : int32 ndarray, optional
+        The calculated distance transform. Returned only when
+        `return_distances` is True, and `distances` is not supplied.
+        It will have the same shape as the input array.
+    indices : int32 ndarray, optional
+        The calculated feature transform. It has an input-shaped array for each 
+        dimension of the input. See distance_transform_edt documentation for an 
+        example.
+        Returned only when `return_indices` is True, and `indices` is not 
+        supplied.
 
     """
     if (not return_distances) and (not return_indices):
@@ -1992,8 +2028,8 @@ def distance_transform_cdt(input, metric='chessboard', return_distances=True,
     else:
         try:
             metric = numpy.asarray(metric)
-        except Exception:
-            raise RuntimeError('invalid metric provided')
+        except Exception as e:
+            raise RuntimeError('invalid metric provided') from e
         for s in metric.shape:
             if s != 3:
                 raise RuntimeError('metric sizes must be equal to 3')
@@ -2012,7 +2048,7 @@ def distance_transform_cdt(input, metric='chessboard', return_distances=True,
 
     rank = dt.ndim
     if return_indices:
-        sz = numpy.product(dt.shape, axis=0)
+        sz = numpy.prod(dt.shape, axis=0)
         ft = numpy.arange(sz, dtype=numpy.int32)
         ft.shape = dt.shape
     else:
@@ -2059,41 +2095,52 @@ def distance_transform_cdt(input, metric='chessboard', return_distances=True,
 def distance_transform_edt(input, sampling=None, return_distances=True,
                            return_indices=False, distances=None, indices=None):
     """
-    Exact euclidean distance transform.
+    Exact Euclidean distance transform.
 
     In addition to the distance transform, the feature transform can
     be calculated. In this case the index of the closest background
-    element is returned along the first axis of the result.
+    element to each foreground element is returned in a separate array.
 
     Parameters
     ----------
     input : array_like
         Input data to transform. Can be any type but will be converted
         into binary: 1 wherever input equates to True, 0 elsewhere.
-    sampling : float or int, or sequence of same, optional
+    sampling : float, or sequence of float, optional
         Spacing of elements along each dimension. If a sequence, must be of
         length equal to the input rank; if a single number, this is used for
         all axes. If not specified, a grid spacing of unity is implied.
     return_distances : bool, optional
-        Whether to return distance matrix. At least one of
-        return_distances/return_indices must be True. Default is True.
+        Whether to calculate the distance transform.
+        Default is True.
     return_indices : bool, optional
-        Whether to return indices matrix. Default is False.
-    distances : ndarray, optional
-        Used for output of distance array, must be of type float64.
-    indices : ndarray, optional
-        Used for output of indices, must be of type int32.
+        Whether to calculate the feature transform.
+        Default is False.
+    distances : float64 ndarray, optional
+        An output array to store the calculated distance transform, instead of 
+        returning it.
+        This parameter is only used when `return_distances` is True.
+        It must be the same shape as `input`.
+    indices : int32 ndarray, optional
+        An output array to store the calculated feature transform, instead of 
+        returning it.
+        Its shape must be `(input.ndim,) + input.shape`.
 
     Returns
     -------
-    distance_transform_edt : ndarray or list of ndarrays
-        Either distance matrix, index matrix, or a list of the two,
-        depending on `return_x` flags and `distance` and `indices`
-        input parameters.
+    distances : float64 ndarray, optional
+        The calculated distance transform. Returned only when
+        `return_distances` is True and `distances` is not supplied.
+        It will have the same shape as the input array.
+    indices : int32 ndarray, optional
+        The calculated feature transform. It has an input-shaped array for each 
+        dimension of the input. See example below.
+        Returned only when `return_indices` is True and `indices` is not 
+        supplied.
 
     Notes
     -----
-    The euclidean distance transform gives values of the euclidean
+    The Euclidean distance transform gives values of the Euclidean
     distance::
 
                     n
@@ -2200,9 +2247,9 @@ def distance_transform_edt(input, sampling=None, return_distances=True,
         if dt_inplace:
             dt = numpy.add.reduce(dt, axis=0)
             if distances.shape != dt.shape:
-                raise RuntimeError('indices has wrong shape')
+                raise RuntimeError('distances has wrong shape')
             if distances.dtype.type != numpy.float64:
-                raise RuntimeError('indices must be of float64 type')
+                raise RuntimeError('distances must be of float64 type')
             numpy.sqrt(dt, distances)
         else:
             dt = numpy.add.reduce(dt, axis=0)
